'''Many core functions for pgeng'''
#IMPORTS
import pygame
from collections import Counter
from sys import exit as _sysexit #UNDERSCORE SO IT IS NOT VISIBLE AS A FUNCTION
#IMPORTS

#CLIP_SURFACE
def clip_surface(surface, location, size):
	'''Creates a new Surface from a part of another Surface

	Returns: pygame.Surface'''
	new_surface = surface.copy()
	new_surface.set_clip(pygame.Rect(location, size))
	clipped_surface = surface.subsurface(new_surface.get_clip())
	return clipped_surface
#CLIP_SURFACE

#LOAD_IMAGE
def load_image(path, colourkey=None, alpha=255):
	'''Load an image for pygame that will be converted
	You can set a colourkey and alpha as well

	Returns: pygame.Surface'''
	image = pygame.image.load(path).convert()
	image.set_colorkey(colourkey)
	if alpha != 255:
		image.set_alpha(alpha)
	return image
#LOAD_IMAGE

#DELTA_TIME
def delta_time(clock, fps):
	'''Get the time since the last frame, it needs a pygame.time.Clock object and the fps
	It will return a number that is around 1 if the game is running at the intended speed
	For example, if the game is running at 30 fps, but it should run at 60 fps, it would return 2.0

	Returns: Float'''
	return clock.get_time() * fps / 1000
#DELTA_TIME

#QUIT_GAME
def quit_game():
	'''Exiting a pygame program in the correct way
	Runs:
		pygame.quit() and sys.exit()'''
	pygame.quit()
	_sysexit()
#QUIT_GAME

#READ_FILE
def read_file(path):
	'''Reads a file and returns the that's data in it

	Returns: String'''
	file = open(path)
	data = file.read()
	file.close()
	return data
#READ_FILE

#WRITE_TO_FILE
def write_to_file(path, data):
	'''Writes data to a file
	The data has to be a string'''
	file = open(path, 'w')
	file.write(data)
	file.close()
#WRITE_TO_FILE

#NEAREST
def nearest(input, nearest, int_mode=True):
	'''Returns a number changed to the nearest given
	If int_mode is True, it will truncate every returned number and return an integer

	Returns: Integer (or float)'''
	return int(round(input / nearest) * nearest) if int_mode else round(input / nearest) * nearest
#NEAREST

#MOST_USED
def most_used(iterable, amount=False):
	'''Return the most used value in an iterable
	It will return the amount used of the value in the list if amount is True

	Returns: List'''
	list_counter = Counter(iterable)
	total_times = list(list_counter.values()).count(max(list(list_counter.values())))
	return list_counter.most_common(total_times) if amount else [value[0] for value in list_counter.most_common(total_times)]
#MOST_USED

#STRING_NUMBER
def string_number(string, return_index=None, int_mode=False):
	'''Return numbers from a given string
	If return_index is None it will return everything, except if a list or tuple with the indexes gets given
	Example:
		return_index=[0, 2] This will make it return the first and third number

	If int_mode is True, it will truncate every returned number and return an integer

	Returns: List (None if no number is found)'''
	if any(character.isdigit() for character in string):
		numbers, digits = [], []
		for i in range(len(string)):
			if string[i].isdigit() and (any(string[min(i + 1, len(string) - 1)] == character for character in ['.', ',']) or string[min(i + 1, len(string) - 1)].isdigit()): #CHECK IF THE NEXT CHARACTER IS . OR , OR A NUMBER
				if any(string[min(i + 1, len(string) - 1)] == character for character in ['.', ',']) and string[min(i + 2, len(string) - 1)].isdigit() and not any('.' in digit for digit in digits): #IF NEXT CHARACTER IS . AND AFTER THAT IS A NUMBER AND A . IS NOT ALREADY IN THE DIGITS
					digits.append(f'{string[i]}.')
				elif any(string[min(i + 1, len(string) - 1)] == character for character in ['.', ',']): #CHECK IF ELSE NUMBERS ENDS WITH A . OR ,
					digits.append(string[i])
					numbers.append(''.join(digits))
					digits = []
				else: #NORMAL SITUATION
					digits.append(string[i])
			elif string[i].isdigit(): #IF ELSE IT'S THE LAST DIGIT IN THE NUMBER
				digits.append(string[i])
				numbers.append(''.join(digits))
				digits = []
			if string[i].isdigit() and i == len(string) - 1: #IF THE NUMBER IS THE LAST CHARACTER OF THE STRING
				numbers.append(''.join(digits))
		if type(return_index) is not list and type(return_index) is not tuple:
			return_index = [i for i in range(len(numbers))]
		if return_index and min(return_index) < -len(numbers) or max(return_index) > len(numbers) - 1:
			raise IndexError('Index in return_index is not possible')
		return [int(float(numbers[index])) for index in return_index] if int_mode else [float(numbers[index]) for index in return_index]
	return None #RETURN NONE IF THERE IS NO NUMBER FOUND IN THE STRING
#STRING_NUMBER


'''A class to make the dipslay/screen
It also has functions to enter and exit fullscreen'''
#IMPORTS
import pygame
from pygame._sdl2.video import Window
#IMPORTS

#SCREEN
class Screen:
	'''A class for the display
	It can also set the display fullscreen on all platforms
	The default flag is pygame.SCALED
	Use GetDisplay to get the display from the class

	For more information about the parameters, go to https://www.pygame.org/docs/ref/display.html#pygame.display.set_mode

	Attributes:

	depth

	display

	flags

	fullscreen_size

	pygame_display

	size

	vsync

	window'''
	#__INIT__
	def __init__(self, size, flags=pygame.SCALED, depth=0, display=0, vsync=1, fullscreen=False):
		'''Initialising the class, it should only be done once'''
		self.fullscreen_size = (pygame.display.Info().current_w, pygame.display.Info().current_h)
		self.size = size
		self.flags = flags
		self.depth = depth
		self.display = display
		self.vsync = vsync
		self.fullscreen = fullscreen
		if fullscreen:
			self.toggle_fullscreen(manual=fullscreen)
		else:
			self.pygame_display = pygame.display.set_mode(size, flags, depth, display, vsync)
		self.window = Window.from_display_module()
	#__INIT__

	#TOGGLE_FULLSCREEN
	def toggle_fullscreen(self, manual=None):
		'''Toggle the display to fullscreen or from fullscreen
		manual can be used if you don't want to toggle it, but set it to fullscreen or exit fullscreen manually
		Than manual has to be True (enter fullscreen) or False (exit Fullscreen)

		Returns: pygame.Surface'''
		self.fullscreen = not self.fullscreen
		if type(manual) is bool:
			self.fullscreen = manual
		if self.fullscreen:
			self.pygame_display = pygame.display.set_mode(self.size, pygame.SCALED | pygame.NOFRAME | pygame.FULLSCREEN, self.depth, self.display, self.vsync)
		else:
			pygame.display.toggle_fullscreen()
			self.pygame_display = pygame.display.set_mode(self.size, self.flags, self.depth, self.display, self.vsync)
			self.window = Window.from_display_module()
			self.window.position = (self.fullscreen_size[0] / 2 - self.window.size[0] / 2, self.fullscreen_size[1] / 2 - self.window.size[1] / 2)
		return self.pygame_display
	#TOGGLE_FULLSCREEN

	#SCREENSHOT
	def screenshot(self):
		'''Making a screenshot of the pygame display
		This only works if this class is used as the main display/screen

		Returns: pygame.Surface'''
		return self.pygame_display.copy()
	#SCREENSHOT

	#GET_DISPLAY
	def get_display(self):
		'''Used to return the pygame display

		Returns: pygame.Surface'''
		return self.pygame_display
	#GET_DISPLAY
#SCREEN


'''Functions and classes for tiles and physics objects'''
#IMPORTS
import pygame
#IMPORTS

#VARIABLES
__all__ = ['tile_size', 'Tile']
tile_size = 0
#VARIABLES

#TILE
class Tile:
	'''A tile used by the Entity object to check collision
	It also supports ramps (45 degrees)
	Ramp 0 = No ramp
	Ramp 1 = top left
	Ramp 2 = top right
	Ramp 3 = bottom right
	Ramp 4 = bottom left

	Attributes:

	location

	size

	ramp'''
	#__INIT__
	def __init__(self, location, size, ramp=0):
		'''Initialising a tile'''
		self.location = list(location)
		self.size = size
		self.ramp = ramp
	#__INIT__

	#RECT
	@property
	def rect(self):
		'''Returns a pygame.Rect object of the tile'''
		return pygame.Rect(self.location[0], self.location[1], self.size, self.size)
	#RECT
#TILE


'''A classes and functions for creating fonts and text buttons'''
#IMPORTS
import pygame
from pathlib import Path
from .Core import clip_surface, load_image
from .Colour import palette_swap
#IMPORTS

#VARIALBES
__all__ = ['create_font', 'TextButton']
path = Path( __file__ ).parent.resolve()
#VARIABLES

#CREATE_FONT
def create_font(colour):
	'''A function to create small and large Font objects
	colour will be the colour of the text
	The first value in the returned tuple is the small font and the second value is the large font

	Returns: Tuple'''
	if tuple(colour) == (0, 0, 0):
		small_font_image = palette_swap(load_image(f'{path}/Font/Small_Font.png'), {(255, 0, 0): colour, tuple(colour): (255, 255, 255)})
		large_font_image = palette_swap(load_image(f'{path}/Font/Large_Font.png'), {(255, 0, 0): colour, tuple(colour): (255, 255, 255)})
		return Font(small_font_image, background_colour=255), Font(large_font_image, background_colour=255)
	if tuple(colour) == (127, 127, 127):
		small_font_image = palette_swap(load_image(f'{path}/Font/Small_Font.png'), {(255, 0, 0): colour, tuple(colour): (128, 128, 128)})
		large_font_image = palette_swap(load_image(f'{path}/Font/Large_Font.png'), {(255, 0, 0): colour, tuple(colour): (128, 128, 128)})
		return Font(small_font_image, 128), Font(large_font_image, 128)
	small_font_image = palette_swap(load_image(f'{path}/Font/Small_Font.png'), {(255, 0, 0): colour})
	large_font_image = palette_swap(load_image(f'{path}/Font/Large_Font.png'), {(255, 0, 0): colour})
	return Font(small_font_image), Font(large_font_image)
#CREATE_FONT

#FONT
class Font:
	'''A class to create a pixel art font
	It will get all the letters out of the image and render them
	The border between letters is usually (127, 127, 127) and the background is usually (0, 0, 0) change them if it is necessary

	The font is made by DaFluffyPotato

	Attributes:

	character_height

	characters

	font_image

	space_width'''
	#__INIT__
	def __init__(self, font_image, border_colour=127, background_colour=0):
		'''Initialising a font object'''
		self.font_image = font_image
		self.font_image.set_colorkey((0, 0, 0) if not background_colour else [background_colour for i in range(3)])
		self.characters = {}
		current_width, character_count = 0, 0
		character_order = ['A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z','a','b','c','d','e','f','g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z','.','-',',',':','+','\'','!','?','0','1','2','3','4','5','6','7','8','9','(',')','/','_','=','\\','[',']','*','"','<','>',';']
		for x in range(self.font_image.get_width()):
			colour = self.font_image.get_at((x, 0))
			if colour[:-1] == (border_colour, border_colour, border_colour): #IF THE TEXT COLOR IS (127, 127, 127), CHANGE BORDER_COLOR
				character_image = clip_surface(self.font_image, (x - current_width, 0), (current_width, self.font_image.get_height())) #CLIP EVERY CHARACTER OUT OF THE FONT IMAGE
				self.characters[character_order[character_count]] = character_image
				character_count += 1
				current_width = 0
			else:
				current_width += 1
		self.space_width, self.character_height = self.characters['A'].get_size()
	#__INIT__

	#GET_SIZE
	def get_size(self, text):
		'''Get the size that that a rendered string would use
		It will return the width and height

		Returns: Tuple'''
		if type(text) is not str:
			raise TypeError(f'{text} is not a string')
		width, height = 0, 0
		for character in text:
			if character not in ('\n', ' ') and character in self.characters:
				width += self.characters[character].get_width() + 1 #+ 1 FOR SPACING
			elif character == ' ' or character not in ['\n']:
				width += self.space_width + 1 #+ 1 FOR SPACING
			else:
				width = 0
				height += self.character_height + 1 #+ 1 FOR SPACING
		return width, height
	#GET_SIZE

	#RENDER
	def render(self, surface, text, location):
		'''Render a string on a surface at a location'''
		if type(text) is not str:
			raise TypeError(f'{text} is not a string')
		x_offset, y_offset = 0, 0
		for character in text:
			if character not in ('\n', ' ') and character in self.characters:
				surface.blit(self.characters[character], (location[0] + x_offset, location[1] + y_offset))
				x_offset += self.characters[character].get_width() + 1 #+ 1 FOR SPACING
			elif character == ' ' or character not in ['\n']:
				x_offset += self.space_width + 1 #+ 1 FOR SPACING
			else:
				x_offset = 0
				y_offset += self.character_height + 1 #+ 1 FOR SPACING
	#RENDER
#FONT

#TEXTBUTTON
class TextButton:
	'''A string of text that is also a button
	The collide function is to collide with the mouse and clicks
	It also needs a font size, it has to be either 'small' or 'large'

	Attributes:

	rect

	test_font

	text

	size'''
	#__INIT__
	def __init__(self, text, location, font_size):
		if font_size != 'small' and font_size != 'large':
			raise ValueError(f'{font_size} is not \'small\' or \'large\'')
		if type(text) is not str:
			raise TypeError(f'{text} is not a string')
		self.text = text
		self.test_font = Font(load_image(f'{path}/Font/{font_size.title()}_Font.png'))
		self.size = self.test_font.get_size(text)
		self.rect = pygame.Rect(location, (self.size[0] - 1, self.size[1] + self.test_font.character_height)) #- 1 FOR THE EXTRA SPACING
	#__INIT__

	#SET_TEXT
	def set_text(self, text):
		'''Sets a new string as the text
		All the variables will be updated, so the functions can be used the same'''
		if type(text) is not str:
			raise TypeError(f'{text} is not a string')
		self.text = text
		self.size = self.test_font.get_size(text)
		self.rect = pygame.Rect(self.rect.topleft, (self.size[0] - 1, self.size[1] + self.test_font.character_height))
	#SET_TEXT

	#COLLIDE
	def collide(self, click):
		'''This will check collision with the mouse location and also if click is True with it
		The first value returns True if the mouse has collided with the button, the second one is if the mouse clicked on it

		Returns: Tuple'''
		if self.rect.collidepoint(pygame.mouse.get_pos()):
			if click:
				return True, True
			return True, False
		return False, False
	#COLLIDE

	#RENDER
	def render(self, surface, font):
		'''Renders the text from the button'''
		if not isinstance(font, Font):
			raise TypeError(f'{font} is not a Font object')
		font.render(surface, self.text, self.rect.topleft)
	#RENDER
#TEXTBUTTON


'''Functions and classes for visual effects in PyGame'''
#IMPORTS
import pygame, math
#IMPORTS

#CIRCLE_LIGHTING
def circle_lighting(radius, colour, alpha=255):
	'''Creates a surface twice as big as the given radius
	It draws a circle with the radius on it
	Recommended to blit with the special_flag pygame.BLEND_RGBA_ADD

	Returns: pygame.Surface'''
	lighting_surface = pygame.Surface((radius * 2, radius * 2))
	lighting_surface.set_colorkey((0, 0, 0))
	if alpha != 255:
		image.set_alpha(alpha)
	pygame.draw.circle(lighting_surface, colour, (radius, radius), radius)
	return lighting_surface
#CIRCLE_LIGHTING

#PARITCLE
class Particle:
	'''A particle to move and render
	It can also have gravity and a 'lighting' for the rendering

	Attributes:

	alive

	colour

	location

	momentum

	size
	'''
	#__INIT__
	def __init__(self, location, momentum, size, colour):
		'''Initialising a Particle object'''
		self.location = list(location)
		self.momentum = momentum
		self.size = size
		self.colour = colour
		self.alive = True
	#__INIT__

	#MOVE
	def move(self, size_change, y_momentum=0, delta_time=1):
		'''Move the location of the Particle and change the size
		If the size is smaller or equal to 0, it will no longer be alive
		It also has gravity (y_momentum)'''
		self.location = list(self.location)
		self.momentum[1] += y_momentum * delta_time
		self.location[0] += self.momentum[0] * delta_time
		self.location[1] += self.momentum[1] * delta_time
		self.size -= size_change * delta_time
		if self.size < 1:
			self.alive = False
	#MOVE

	#RENDER
	def render(self, surface, scroll=(0, 0), lighting_colour=None):
		'''Render the Particle if it is alive
		scroll is position of the camera, it will render it at the location of the Particle minus the scroll
		It can also render a lighting circle around the particle with the pygame.BLEND_RGBA_ADD flag'''
		if self.alive:
			pygame.draw.circle(surface, self.colour, (self.location[0] - scroll[0], self.location[1] - scroll[1]), self.size)
			if lighting_colour:
				lighting_radius = self.size * 2
				surface.blit(circle_lighting(lighting_radius, lighting_colour), (self.location[0] - lighting_radius - scroll[0], self.location[1] - lighting_radius - scroll[1]), special_flags=pygame.BLEND_RGBA_ADD)
	#RENDER
#PARTICLE

#SHOCKWAVE
class Shockwave:
	'''A shockwave that gets smaller or bigger

	Attributes:

	alive

	colour

	location

	radius

	width'''
	#__INIT__
	def __init__(self, location, radius, width, colour):
		'''Initialising a Shockwave object'''
		self.location = list(location)
		self.radius = radius
		self.width = width
		self.colour = colour
		self.alive = True
	#__INIT__

	#UPDATE
	def update(self, surface, radius_change, width_change, scroll=(0, 0), delta_time=1):
		'''The move and render function of the Shockwave (it will only render it if it is still alive)
		If the width is smaller or the same as 1, it will no longer be alive
		scroll is position of the camera, it will render it at the location of the Shockwave minus the scroll'''
		self.radius += radius_change * delta_time
		self.width -= width_change * delta_time
		if self.width <= 1:
			self.alive = False
		else:
			pygame.draw.circle(surface, self.colour, (self.location[0] - scroll[0], self.location[1] - scroll[1]), self.radius, int(self.width))
	#UPDATE
#SHOCKWAVE

#SPARK
class Spark:
	'''A spark to draw and move
	A spark is a polygon so checking collision with it is not possible'''
	#__INIT__
	def __init__(self, location, angle, speed, size, colour):
		'''Initialising a Spark object
		It will change the angle to radians

		Attributes:

		alive

		angle

		colour

		location

		size

		speed'''
		self.location = list(location)
		self.angle = math.radians(angle) #ANGLE IS IN RADIANS
		self.speed = speed
		self.size = size
		self.colour = colour
		self.alive = True
	#__INIT__

	#NORMAL_MOVEMENT
	def normal_movement(self, speed_change, delta_time=1, angle_change=0):
		'''A function to move the Spark
		It will decrease the speed and if it is lower or equal to 0, it is no longer alive
		angle_change will change the angle by that amount every time this function runs'''
		self.location = list(self.location)
		self.location[0] += math.cos(self.angle) * self.speed * delta_time
		self.location[1] += math.sin(self.angle) * self.speed * delta_time
		self.speed -= speed_change * delta_time
		self.angle += angle_change * delta_time
		if self.speed <= 0:
			self.alive = False
	#NORMAL_MOVEMENT

	#GRAVITY_MOVEMENT
	def gravity_movement(self, speed_change, y_momentum, maximum_y_momentum, delta_time=1):
		'''A function to move the Spark, but it also includes gravity
		It will change the angle a small amount down every frame, that will be set with y_momentum
		It will decrease the speed and if it is lower or equal to 0, it is no longer alive'''
		self.location = list(self.location)
		momentum = [math.cos(self.angle) * self.speed * delta_time, math.sin(self.angle) * self.speed * delta_time]
		self.location[0] += momentum[0]
		self.location[1] += momentum[1]
		momentum[1] = min(momentum[1] + y_momentum * delta_time, maximum_y_momentum)
		self.angle = math.atan2(momentum[1], momentum[0])
		self.speed -= speed_change * delta_time
		if self.speed <= 0:
			self.alive = False
	#GRAVITY_MOVEMENT

	#RENDER
	def render(self, surface, scroll=(0, 0), lighting_colour=None):
		'''A function to render the Spark if it is alive
		It will calculate every point for the polygon (there are 4 points) that looks like a rhombus, but the back one is longer
		scroll is position of the camera, it will render it at the location of the Spark minus scroll
		If lighting_colour is turned on, it will blit a bigger polygon with a specific colour on top of it with BLEND_RGBA_ADD'''
		if self.alive:
			points = [[self.location[0] + math.cos(self.angle) * self.speed * self.size, self.location[1] + math.sin(self.angle) * self.speed * self.size], #FRONT POINT
			[self.location[0] + math.cos(self.angle + math.pi / 2) * self.speed * self.size * 0.4, self.location[1] + math.sin(self.angle + math.pi / 2) * self.speed * self.size * 0.4], #RIGHT POINT
			[self.location[0] - math.cos(self.angle) * self.speed * self.size * 2.5, self.location[1] - math.sin(self.angle) * self.speed * self.size * 2.5], #BOTTOM POINT
			[self.location[0] + math.cos(self.angle - math.pi / 2) * self.speed * self.size * 0.4, self.location[1] + math.sin(self.angle - math.pi / 2) * self.speed * self.size * 0.4]] #LEFT POINT

			pygame.draw.polygon(surface, self.colour, [[point[0] - scroll[0], point[1] - scroll[1]] for point in points])
			if lighting_colour:
				lighting_position = [abs(min([point[0] for point in points]) - self.location[0] + (min([point[0] for point in points]) - max([point[0] for point in points])) * 0.5),
				abs(min([point[1] for point in points]) - self.location[1] + (min([point[1] for point in points]) - max([point[1] for point in points])) * 0.5)]
				surface.blit(self.lighting(points, lighting_position, lighting_colour), (self.location[0] - lighting_position[0] - scroll[0], self.location[1] - lighting_position[1] - scroll[1]), special_flags=pygame.BLEND_RGBA_ADD)
	#RENDER

	#LIGHTING
	def lighting(self, points, lighting_position, colour):
		'''A function used by render to draw a bigger polygon on top of the normal one'''
		lighting_surface_width = (max([point[0] for point in points]) - min([point[0] for point in points])) * 2
		lighting_surface_height = (max([point[1] for point in points]) - min([point[1] for point in points])) * 2
		lighting_surface = pygame.Surface((lighting_surface_width, lighting_surface_height)) #DIFFERENCE BETWEEN LOWEST AND HIGHTEST POINTS
		lighting_surface.set_colorkey((0, 0, 0))

		#EVERYTHING * 2 FOR MAKING THE POLYGON LARGER
		larger_points = [[lighting_position[0] + math.cos(self.angle) * self.speed * self.size * 2, lighting_position[1] + math.sin(self.angle) * self.speed * self.size * 2], #FRONT POINT
		[lighting_position[0] + math.cos(self.angle + math.pi / 2) * self.speed * self.size * 0.8, lighting_position[1] + math.sin(self.angle + math.pi / 2) * self.speed * self.size * 0.8], #RIGHT POINT
		[lighting_position[0] - math.cos(self.angle) * self.speed * self.size * 5, lighting_position[1] - math.sin(self.angle) * self.speed * self.size * 5], #BOTTOM POINT
		[lighting_position[0] + math.cos(self.angle - math.pi / 2) * self.speed * self.size * 0.8, lighting_position[1] + math.sin(self.angle - math.pi / 2) * self.speed * self.size * 0.8]] #LEFT POINT
		pygame.draw.polygon(lighting_surface, colour, larger_points)
		return lighting_surface
		#LIGHTING
#SPARK


'''Functions used for changing colours of surfaces'''
#IMPORTS
import pygame
from .Core import nearest
#IMPORT

#PALETTE_SWAP
def palette_swap(surface, colours):
	'''Swaps the colours of a pygame.Surface
	colours = {old_colour: new_colour, old_colour: new_colour}

	Returns: pygame.Surface'''
	new_surface = surface.copy()
	for y in range(new_surface.get_height()):
		for x in range(new_surface.get_width()):
			colour_value = new_surface.get_at((x, y))
			if colour_value[:-1] in colours:
				new_surface.set_at((x, y), colours[colour_value[:-1]])
	return new_surface
#PALETTE_SWAP

#GRAY_SCALE
def gray_scale(surface):
	'''A basic algorithm for grayscaling a surface (Luma ITU-R Recommendation BT. 709)
	Formula: Red * 0.2126 + Green * 0.7152 + Blue * 0.0722

	Returns: pygame.Surface'''
	new_surface = surface.copy()
	for y in range(new_surface.get_height()):
		for x in range(new_surface.get_width()):
			colour_value = new_surface.get_at((x, y))
			new_surface.set_at((x, y), [colour_value[0] * 0.2126 + colour_value[1] * 0.7152 + colour_value[2] * 0.0722 for i in range(3)])
	return new_surface
#GRAY_SCALE

#GRAY_SHADE
def gray_shade(surface, shades=16):
	'''A basic algorithm for grayscaling a surface with shades
	It will change the colour to one of the shades
	The amount of shades must be between 2 and 256

	Returns: pygame.Surface'''
	if not 2 < shades < 256:
		raise ValueError(f'{shades} Shades not supported, the amount must be between 2 and 256')
	conversion = 255 / (shades - 1)
	new_surface = surface.copy()
	for y in range(new_surface.get_height()):
		for x in range(new_surface.get_width()):
			colour_value = new_surface.get_at((x, y))
			new_surface.set_at((x, y), [nearest((colour_value[0] + colour_value[1] + colour_value[2]) / 3, conversion) for i in range(3)])
	return new_surface
#GRAY_SHADE


'''Useful functions and classes for PyGame'''
#VARIABLES
failed_imports = []
#VARIABLES

#ANIMATIONS
try:
	from .Animations import *
except ImportError as e:
	failed_imports.append(e.name)
#ANIMATIONS

#COLOUR
try:
	from .Colour import *
except ImportError as e:
	failed_imports.append(e.name)
#COLOUR

#CORE
try:
	from .Core import *
except ImportError as e:
	failed_imports.append(e.name)
#CORE

#ENTITY
try:
	from .Entity import Entity
except ImportError as e:
	failed_imports.append(e.name)
#ENTITY

#FONT
try:
	from .Font import *
except ImportError as e:
	failed_imports.append(e.name)
#FONT

#SCREEN
try:
	from .Screen import Screen
except ImportError as e:
	failed_imports.append(e.name)
#SCREEN

#TILE
try:
	from .Tile import *
except ImportError as e:
	failed_imports.append(e.name)
#TILE

#SOUNDS
try:
	from .Sounds import Sounds
except ImportError as e:
	failed_imports.append(e.name)
#SOUNDS

#VFX
try:
	from .VFX import *
except ImportError as e:
	failed_imports.append(e.name)
#VFX

#MESSAGES
print('''\nDo help(pgeng.<Function>) to get the documentation of the function (or possibly class)
Or do help(pgeng.<Package Content>) to get the documentation of the functions or classes in a file\n''')
print(f'Failed to import {", ".join([failed_import.split(".")[1] for failed_import in failed_imports])} from pgeng') if failed_imports else None
del failed_imports
#MESSAGES


'''A class for animations'''
#IMPORTS
import pygame
from pathlib import Path
from .Core import load_image, most_used
#IMPORTS

#ANIMATIONS
class Animations:
	'''A class for animations
	Not all the animations should be done in one Animations class, it is good to create one for Entities and such
	It needs a action to begin with

	Attributes:

	action

	animation_frames

	animations_data

	frame'''
	#__INIT__
	def __init__(self, action):
		'''Initialising an Animations class'''
		self.animations_data = {}
		self.animation_frames = {}
		self.action = action
		self.frame = 0
	#__INIT__

	#ADD_ANIMATION
	def load_animation(self, path, frame_durations, repeat=True, colourkey=None, file_type=None, animation_name=None):
		'''A function for loading still images for a animation

		It loads all images in a directory, the last character before the file type needs to be a number
		The name of the directory it is in needs to be in the file name (directories above that don't matter),
		The file type of the frames should be the most used file type in the directory
		Example:
			'run/run1.png'
			'run/run2.png'

		Or you can use file_type variable
		Example:
			file_type='.png'

		If you don't want the name of the directory to be the name of the animation, you could use animation_name
		If you do that, the file name should still be the same as the directory

		frame_durations is a list or tuple with how long each frame should last, it has to be in order
		repeat is if the animation should loop, if it is False, the image would stay on the last one
		The images will be added to animation_frames'''
		path = Path(path)
		if not path.is_dir():
			raise FileNotFoundError(f'Directory \'{path}\' does not exist')
		animation_name = path.name if animation_name is None else animation_name
		self.animations_data[animation_name] = ([], repeat)
		if not file_type:
			file_type = most_used([file.suffix for file in path.glob('*.*')])[0]
		for i, frame_duration in enumerate(frame_durations):
			if not path.joinpath(f'{path.name}{i + 1}{file_type}').is_file():
				raise pygame.error(f'Too many non image files in \'{path}\' or the image file is not named \'{path.name}{i + 1}{file_type}\'')
			animation_frame_id = f'{animation_name}{i + 1}'
			self.animation_frames[animation_frame_id] = load_image(path.joinpath(f'{path.name}{i + 1}{file_type}'), colourkey)
			self.animations_data[animation_name][0].append((animation_frame_id, frame_duration))
	#ADD_ANIMATION

	#ADD_IMAGE
	def add_image(self, image, name, duration, repeat=True):
		'''Manually add an image to an animation
		If this name does not yes exist, it will be created
		The image will be added to animation_frames with a number, so it has to be added in the correct order'''
		if name not in self.animations_data:
			self.animations_data[name] = ([], repeat)
		animation_frame_id = f'{name}{len(self.animations_data[name][0]) + 1}'
		self.animation_frames[animation_frame_id] = image
		self.animations_data[name][0].append((animation_frame_id, duration))
	#ADD_IMAGE

	#CURRENT_IMAGE
	def current_image(self, delta_time=1):
		'''Get the current image in the animation
		delta_time is how much the frame should update, usually it would be 1

		Returns: pygame.Surface'''
		if self.action not in self.animations_data:
			raise KeyError(f'Animation \'{action}\' is not defined')
		animation_data = self.animations_data[self.action]
		self.frame += delta_time
		reset_frame = round(self.frame) > sum([frame_duration[1] for frame_duration in animation_data[0]][:len(animation_data[0])])
		if reset_frame and animation_data[1]:
			self.frame = 0
		elif reset_frame and not animation_data[1]:
			self.frame = sum([frame_duration[1] for frame_duration in animation_data[0]][:len(animation_data[0])])
		for i, frame_data in enumerate(animation_data[0]):
			if sum([frame_duration[1] for frame_duration in animation_data[0]][:i + 1]) >= round(self.frame):
				return self.animation_frames[frame_data[0]]
	#CURRENT_IMAGE

	#SET_ACTION
	def set_action(self, action):
		'''Set the current action, the frame will also be reset to 0'''
		if action not in self.animations_data:
			raise KeyError(f'Animation \'{action}\' is not defined')
		if self.action != action:
			self.action = action
			self.frame = 0
	#SET_ACTION
#ANIMATIONS


'''A simple class for loading and playing sounds and music'''
#IMPORTS
import pygame
#IMPORTS

#SOUNDS
class Sounds:
	'''A class for playing sounds or music
	It uses a boolean variable to know if it should play sound
	It will not if PlaySoundVariable is False
	The volume will be set to 1 automatically

	Channels is for setting the amount of channels in pygame.mixer if it is initialised

	Attributes:

	play_sound_variable

	sounds

	volume'''
	#__INIT__
	def __init__(self, channels=64):
		'''Initialising the class, it should only be done once'''
		if pygame.mixer.get_init():
			pygame.mixer.set_num_channels(channels)
		self.sounds = {}
		self.play_sound_variable = True
		self.volume = 1
	#__INIT__

	#ADD_SOUND
	def add_sound(self, path, name, volume=-1):
		'''Adding sounds to the Sounds dictionary
		The volume will be set by the volume variable if volume is -1'''
		self.sounds[name] = pygame.mixer.Sound(path)
		self.sounds[name].set_volume(self.volume if volume == -1 else volume)
	#ADD_SOUND

	#SET_PLAY_SOUND
	def set_play_sound(self, play_sound_boolean, affect_music=True):
		'''Setting the play_sound_variable to True or False
		It will also pause music if it is False and unpause if it is True'''
		self.play_sound_variable = play_sound_boolean
		if self.play_sound_variable and not pygame.mixer.get_busy() and affect_music:
			pygame.mixer.music.unpause()
		elif affect_music:
			pygame.mixer.music.pause()
	#SET_PLAY_SOUND

	#SET_VOLUME
	def set_volume(self, volume, music=True, specific_sound=None):
		'''Set the volume of the sounds
		It also sets the volume variable for new sounds in the future

		It will not change the volume of the music if music is False
		It will set the volume of a specific sound in the dictionary if a specific_sound is given
		Than it will not change the volume variable'''
		if 0 > volume > 1:
			raise ValueError('Volume must be between 0 and 1')
		if music:
			pygame.mixer.music.set_volume(volume)
		if specific_sound is not None:
			self.volume = volume
			for sound in self.sounds:
				self.sounds[sound].set_volume(volume)
		else:
			self.sounds[specific_sound].set_volume(volume)
	#SET_VOLUME

	#PLAY_SOUND
	def play_sound(self, sound):
		'''Plays the specified sound in the sounds dictionary if play_sound_variable is True'''
		if self.play_sound_variable:
			self.sounds[Sound].play()
	#PLAY_SOUND

	#PLAY_MUSIC
	def play_music(self, path, amount=-1, volume=-1):
		'''This will play the music file if play_sound_variable is True
		The volume will be the volume variable if volume is -1
		amount is the amount of times played, it will be indefinitely if amount is -1'''
		if self.play_sound_variable:
			pygame.mixer.music.load(path)
			pygame.mixer.music.set_volume(self.volume if volume == -1 else volume)
			pygame.mixer.music.play(amount)
	#PLAY_MUSIC
#SOUNDS


'''An Entity class'''
#IMPORTS
import pygame, math
from .Animations import Animations
#IMPORTS

#COLLISION_TEST
def collision_test(rect, collision_list):
	'''A function used by the Entity class to check collisions with tiles'''
	return [list_rect for list_rect in collision_list if rect.colliderect(list_rect)]
#COLLISION_TEST

#ENTITY
class Entity:
	'''An entity with usefull functions
	It needs a position and a size for the pygame.Rect object
	Use Entity.location[0] and Entity.location[1] instead of Entity.rect.x and Entity.rect.y
	It has an animation class

	Attributes:

	alpha

	animations

	flips

	location

	rect

	rotation

	scale'''
	#__INIT__
	def __init__(self, location, size):
		'''Initialising an Entity object'''
		self.rect = pygame.Rect(location, size)
		self.location = list(location)

		self.flips = [False, False]
		self.rotation = 0
		self.scale = [1, 1]
		self.alpha = 255

		self.animations = Animations(None)
	#__INIT__

	#CENTER
	@property
	def center(self):
		'''Returns the center location of the Entity'''
		return [self.location[0] + self.rect.w / 2, self.location[1] + self.rect.h / 2]
	#CENTER

	#GET_ANGLE
	def get_angle(self, target):
		'''Get the angle to a target
		target can be an Entity object or a list or tuple with an x and y coordinate
		It will go from the center of the Entity
		It will returns the angle in radians, not degrees

		Returns: Float'''
		if isinstance(target, Entity):
			return math.atan2(target.center[1] - self.center[1], target.center[0] - self.center[0])
		return math.atan2(target[1] - self.center[1], target[0] - self.center[0])
	#GET_ANGLE

	#GET_DISTANCE
	def get_distance(self, target):
		'''Get the distance to a target
		target can be an Entity object or a list or tuple with an x and y coordinate

		Returns: Float'''
		if isinstance(target, Entity):
			return math.sqrt((target.location[0] - self.location[0]) ** 2 + (target.location[1] - self.location[1]) ** 2)
		return math.sqrt((target[0] - self.location[0]) ** 2 + (target[1] - self.location[1]) ** 2)
	#GET_DISTANCE

	#IN_RANGE
	def in_range(self, target, range):
		'''This will check if a target is in a specified range
		target can be an Entity object or a list or tuple with an x and y coordinate
		This uses get_distance

		Returns: Boolean'''
		return self.get_distance(target) <= range
	#IN_RANGE

	#SET_SCALE
	def set_scale(self, scale):
		'''This sets the scale for the image
		It can be a list or tuple with the scale for the width and the scale for the height
		If it is a number, than it will be set for the width and height'''
		if type(scale) is list or type(scale) is tuple:
			self.scale = list(scale)
		else:
			self.scale = [scale, scale]
	#SET_SCALE

	#IMAGE
	def current_image(self, delta_time=1):
		'''Get the current image in the animation
		delta_time is how much the frame should update, usually it would be 1
		It also get transformed with transform_image

		Returns: pygame.Surface'''
		return self.transform_image(self.animations.current_image(delta_time))
	#IMAGE

	#TRANSFORM_IMAGE
	def transform_image(self, image):
		'''This will transform the image with the current values of the variables

		It will scale it according to the scale variable
		It will flip it horizontal and vertical according to the flips variable
		It will rotate it according to the rotation variable, pygame will rotate it in the other direction compared to degrees in math
		It will also set and alpha according to the alpha variable

		For more information about the rotation, go to https://www.pygame.org/docs/ref/transform.html#pygame.transform.rotate

		Returns: pygame.Surface'''
		image = image.copy()
		if self.scale != [1, 1]:
			image = pygame.transform.scale(image, (int(image.get_width() * self.scale[0]), int(image.get_height() * self.scale[1])))
		if any(self.flips):
			image = pygame.transform.flip(image, self.flips[0], self.flips[1])
		if self.rotation:
			image = pygame.transform.rotate(image, self.rotation)
		if self.alpha != 255:
			image.set_alpha(self.alpha)
		return image
	#TRANSFORM_IMAGE

	#MOVEMENT
	def movement(self, momentum, tiles):
		'''Changing the position of the Entity object
		It needs a list that has the location change since last frame [x, y] not the actual x and y coordinates
		It also needs a list with Tile objects in it

		It returns a dictionary with booleans to show what part of the rect collide with the tiles:
		{'top': False, 'bottom': False, 'right': False, 'left': False}

		Returns: Dictionary'''
		#VARIABLES
		self.location = list(self.location)
		collision_types = {'top': False, 'bottom': False, 'right': False, 'left': False}
		normal_tiles = [tile.rect for tile in tiles if not tile.ramp]
		ramp_tiles = [tile for tile in tiles if tile.ramp]
		#VARIABLES

		#NORMAL_TILES
		self.location[0] += momentum[0]
		self.rect.x = self.location[0]
		hit_list = collision_test(self.rect, normal_tiles)
		for tile in hit_list:
			if momentum[0] > 0:
				self.rect.right = tile.left
				collision_types['right'] = True
			elif momentum[0] < 0:
				self.rect.left = tile.right
				collision_types['left'] = True
			self.location[0] = self.rect.x

		self.location[1] += momentum[1]
		self.rect.y = self.location[1]
		hit_list = collision_test(self.rect, normal_tiles)
		for tile in hit_list:
			if momentum[1] > 0:
				self.rect.bottom = tile.top
				collision_types['bottom'] = True
			elif momentum[1] < 0:
				self.rect.top = tile.bottom
				collision_types['top'] = True
			self.location[1] = self.rect.y
		#NORMAL_TILES

		#RAMPS
		for ramp in ramp_tiles:
			ramp_hitbox = ramp.rect
			if self.rect.colliderect(ramp_hitbox):
				delta_x = self.rect.x - ramp_hitbox.x
				if ramp.ramp == 1 or ramp.ramp == 4:
					height_position = ramp.size - delta_x
				else:
					height_position = delta_x + self.rect.w

				height_position = min(height_position, ramp.size)
				height_position = max(height_position, 0)

				if ramp.ramp == 1 or ramp.ramp == 2:
					y = ramp_hitbox.y + height_position
				else:
					y = ramp_hitbox.y + ramp.size - height_position

				if (ramp.ramp == 3 or ramp.ramp == 4) and self.rect.bottom > y:
					self.rect.bottom = y
					self.location[1] = self.rect.y
					collision_types['bottom'] = True
				elif (ramp.ramp == 1 or ramp.ramp == 2) and self.rect.top < y:
					self.rect.top = y
					self.location[1] = self.rect.y
					collision_types['top'] = True
		#RAMPS

		return collision_types
	#MOVEMENT
#ENTITY


import setuptools

with open('README.md', 'r', encoding='utf-8') as fh:
	long_description = fh.read()

setuptools.setup(
	name='pgeng',
	version='1.2',
	author='Qamynn',
	description='Useful functions and classes for PyGame',
	long_description=long_description,
	long_description_content_type='text/markdown',
	url='https://github.com/Qamynn/pgeng',
	license='MIT',
	classifiers=[
		'Programming Language :: Python :: 3',
		'License :: OSI Approved :: MIT License',
		'Operating System :: OS Independent',
	],
	packages=['pgeng'],
	include_package_data=True,
	install_requires=['pygame>=2'],
	python_requires='>=3.6',
)
